/*
 * Copyright (C) 2005 Cesar Kallas
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * E-mail: cesarkallas@gmx.net
 * http://cesarkallas.soulivre.org
 *
 */
 
package estella.cliente;
import  estella.lib.*;
import  estella.gui.*;

import java.io.File;
import java.io.FileReader;
import java.io.BufferedReader;
import java.util.StringTokenizer;
import java.util.Vector;

/**
* Funcoes do cliente
* @author Cesar Kallas
* @since 21/11/2005
*/
public class Cliente 
{
	public javax.swing.JFrame jFrame;
	private Constantes CON;
	private Log LOG;
  private Tempo TEMPO;
  private StringTokenizer preServents;
	private Payload payload;
	private Descritor descritor;
	private Thread t_recebeDescritores;

	/**
	* Tenta conecta aos servents declarados na lista de servents pre-definidas em um arquivo
	* @return true ou false
	*/
  public boolean conecta(String host, int porta)
  {
		try {
			// Tenta conectar, verifica resposta do servidor e se ok, adiciona a lista de conexoes
			Conexao conexao = new Conexao(host, porta);
			if(conexao.inicia()) {
				conexao.setTimeout(CON.SOCKET_TIMEOUT);
				conexao.envia(CON.GNUTELLA);
				if(conexao.recebe().indexOf(CON.GNUTELLA_OK) != -1) {
					conexao.setTimeout(0);
					Servent offlineServent = Dados.offlineServents.remove(host, porta);
					Servent servent;
					if(offlineServent != null)
						servent = new Servent(conexao, offlineServent.numArqComp, offlineServent.tamArqComp, offlineServent.ultVerificacao);
					else
						servent = new Servent(conexao, 0, 0, "desconhecido");
					
					Dados.onlineServents.adiciona(servent);
					Dados.offlineServents.atualizaTabelaServents();
					Dados.onlineServents.atualizaTabelaServentsOnline();					
					
					t_recebeDescritores = new Thread(new recebeDescritores(conexao));
					t_recebeDescritores.start();
					return true;
				}
			}
			LOG.falha(" ao conectar no host " + host + ":" + porta);
			return false;
		}
		catch(Exception FalhaConectar) {
			if(FalhaConectar.getMessage() != null)
				LOG.falha(" ao conectar no host " + host + ":" + porta + " ->" + FalhaConectar.getMessage());
			return false;
		}
	}
	
	/**
	* Desconecta o  servent que tiver o ip igual a host
	* @param host  String com o ip do servent
	* @param porta inteiro com a porta
	* @return true ou false
	*/
	public boolean desconecta(String host, int porta)
	{
		int tamArqComp, numArqComp;
		
		try {
			Servent servent = Dados.onlineServents.remove(host, porta);
			if(servent != null)
				Dados.offlineServents.adiciona(new Servent(host, porta, servent.numArqComp, servent.tamArqComp, servent.ultVerificacao));
			else
				Dados.offlineServents.adiciona(new Servent(host, porta, 0, 0, "desconhecido"));
			
			Dados.offlineServents.atualizaTabelaServents();
			Dados.onlineServents.atualizaTabelaServentsOnline();
			return true;
		}
		catch(Exception desconectaHost) {
			LOG.falha(" desconectando host " + host);
			return false;
		}
	}
		
	public boolean ping()
	{
		try {
			payload = new Payload();
			descritor = new Descritor(CON.PING, CON.TTL, 0, payload.ping());
			if(Dados.onlineServents.envia(new String(descritor.retorna())) == false)
				return false;
			if(Dados.pingEnviados.adiciona(descritor) == false)
				return false;
			return true;
		}
		catch(Exception falhaPing) {
			LOG.info(" ao enviar ping " + falhaPing.getMessage());
			return false;
		}
	}
	
	public boolean query(int minVelocidade, String busca)
	{
		if(busca.length() <=0) {
			LOG.falha(" ao fazer query: sem termo de busca");
			return false;
		}
		try {
			payload = new Payload();
			descritor = new Descritor(CON.QUERY, CON.TTL, 0, payload.query(minVelocidade, busca));
			if(!Dados.onlineServents.envia(new String(descritor.retorna()))) {
				System.out.println(" ao enviar query para servents online");
				return false;
			}
			
			if(!Dados.queryEnviados.adiciona(descritor)) {
				System.out.println(" ao adicionar query na lista de enviadas");
				return false;
			}
			System.out.println(" enviada query("+ descritor.retorna().length + ") termo(" + busca + ")");
			return true;
		}
		catch(Exception falhaQuery) {
			LOG.info(" ao enviar query  " + falhaQuery.getMessage());
			return false;
		}
	}

	/**
	* Retorna uma lista de servents pre-definidas em um arquivo
	* @param caminho_servents String com o caminho para o arquivo com os servents. Arquivo definido no formato --  'servent:porta' --
	* @return String com todos os servents e portas encontrados no arquivo
	*/
	public boolean carregaServents(String caminho_servents)
	{
		BufferedReader lista_servents;
		String linha, servents;
		int i=0;
		
		try {
			lista_servents = new BufferedReader(new FileReader(new File(caminho_servents)));
			linha = new String();
			
			while(linha != null) {
				linha = lista_servents.readLine();
				if(linha == null)
					return true;
				Vector tabelaServent = new Vector();
				StringTokenizer st = new StringTokenizer(linha, ":");
				Servent servent = new Servent(st.nextToken(), Integer.parseInt(st.nextToken()), 0, 0, "desconhecido");
				Dados.offlineServents.adiciona(servent);
			}
			lista_servents.close();
			return true;
		}
		catch(Exception LerServents) {
			LOG.falha(LerServents.getMessage());
			return false;
		}
	}
}

/**
* Thread para tratar requisicao recebida pelo servidor
* @param sck_cliente Socket com a conexao do cliente
* @since 23/11/2005
* @author Cesar Kallas - cesarkallas at gmx . net
*/
class recebeDescritores implements Runnable 
{
  private Conexao conexao;
  private Log LOG;
  private Constantes CON;
	private Descritor descritor;
	private String mensagem;
	private Dados dados;
	private Thread t_trataDescritor;
		
  public recebeDescritores(Conexao cliente)
  {
    LOG = new Log("Cliente");
    CON = new Constantes();
		conexao = cliente;
  }

  public void run()
  {
		LOG.info("Escutando descritores de " + conexao.status());
		// Tratar descritores que cheguem ao cliente
		while(conexao.status() != null) {
			try {
				if(conexao.status().isClosed()) {
					conexao.fecha();
					return;
				}
				mensagem = conexao.recebe();
			}
			catch(Exception ReceberDescritor)
			{
				LOG.falha("" + ReceberDescritor.getMessage());
				return;
			}
			descritor = new Descritor(mensagem);
			t_trataDescritor = new Thread(new trataDescritor(descritor, conexao));
			t_trataDescritor.start();
		}
		return;
	}
}

class trataDescritor implements Runnable
{
	private Descritor descritor;
	private Log LOG;
	private Constantes CON;
	private Conexao conexao;
	private Payload payload;
		
	public trataDescritor(Descritor descritor, Conexao conexao)
	{
		this.descritor = descritor;
		this.conexao = conexao;
		this.CON = new Constantes();
		this.LOG = new Log("Cliente_trataDescritor");
	}
	
	public void run()
	{
		// Tamanho minimo de descritor 20 (id + ttl + hops + payload length)
		if(descritor.retorna().length < 20)
			return;
		
		switch((int)descritor.tipo())
		{
			case (byte)00000000: break;
			case (byte)00000001: { pong(); break; }
			default: { queryhits(); break; }
		}
	}
	
	/** 
	* Trata pongs recebidos pelo cliente
	*/
	private void pong()
	{	
		if(!Dados.pingEnviados.existe(descritor.stringId())) {
			LOG.atencao(" pong recebido mas no localizado na lista de pings enviados id(" + descritor.stringId() + ")");
			return;
		}

		LOG.info("pong recebido id(" + descritor.stringId() + ") ttl(" + descritor.ttl() + ") hops(" + descritor.hops() +  ") <- " + conexao.status());
		Tempo tempo = new Tempo();
		Payload payload = new Payload(CON.PONG, descritor.payload());
		Servent servent = new Servent(payload.pongIp(), payload.pongPorta(), payload.pongNumArqComp(), payload.pongTamArqComp(), tempo.data()+"  "+tempo.hora());
		
		if(Dados.onlineServents.existe(payload.pongIp())) {
			Dados.onlineServents.atualiza(payload.pongIp(), payload.pongNumArqComp(), payload.pongTamArqComp(), tempo.data()+"  "+tempo.hora());
			Dados.onlineServents.atualizaTabelaServentsOnline();
			LOG.info(" atualizada lista de servents online " + payload.pongIp() + " porta: " + payload.pongPorta() + " numArqComp: " + payload.pongNumArqComp() + " tamArqComp: " + payload.pongTamArqComp());
		}
		else {
			if(Dados.offlineServents.existe(payload.pongIp())) {
				Dados.offlineServents.atualiza(servent);
				LOG.info(" atualizada lista de servents offline " + payload.pongIp() + " porta: " + payload.pongPorta() + " numArqComp: " + payload.pongNumArqComp() + " tamArqComp: " + payload.pongTamArqComp());
			}else {
				Dados.offlineServents.adiciona(servent);
				LOG.info(" novo servent encontrado, adicionado na lista de servents offline " + payload.pongIp() + " porta: " + payload.pongPorta() + " numArqComp: " + payload.pongNumArqComp() + " tamArqComp: " + payload.pongTamArqComp());
			}
			Dados.offlineServents.atualizaTabelaServents();
		}
	}
	
	/**
	* Trata os queryhits recebidos pelo cliente
	*/
	private void queryhits()
	{
		if(!Dados.queryEnviados.existe(descritor.stringId())) {
			LOG.atencao(" queryhits recebido mas no localizado na lista de querys enviadas id(" + descritor.stringId() + ")");
			return;
		}
		
		LOG.info("recebido queryhits(" + descritor.retorna().length + ") id(" + descritor.stringId() + ") ttl(" + descritor.ttl() + ") hops(" + descritor.hops() +  ") <- " + conexao.status());
		Payload payload = new Payload(CON.QUERYHITS, descritor.payload());
		
		LOG.info("resultado da pesquisa - hits: " + payload.queryhitsNum() + " ip: " + payload.queryhitsIp() + " porta: " + payload.queryhitsPorta());

		StringTokenizer hits = new StringTokenizer(payload.retornaHit(payload.queryhitsNum(), payload.queryhitsResultset()), CON.SEPARADOR);
		Vector resultadoQueryHits = new Vector();
		
		while(hits.hasMoreTokens()) {
			resultadoQueryHits.addElement(hits.nextToken());
			resultadoQueryHits.addElement(hits.nextToken());
			resultadoQueryHits.addElement(hits.nextToken());
			resultadoQueryHits.addElement(payload.queryhitsIp());
			resultadoQueryHits.addElement(payload.queryhitsPorta() + "");
			synchronized(JanelaPrincipal.linhaTabelaBusca) {
				JanelaPrincipal.linhaTabelaBusca.addElement(resultadoQueryHits);
			}
			resultadoQueryHits = new Vector();
		}
		synchronized(JanelaPrincipal.tabelaBusca) {
			JanelaPrincipal.tabelaBusca.removeNotify();
			JanelaPrincipal.tabelaBusca.addNotify();
		}
	}
}
